import numpy as np

from .function import Function


class ConstantPadding(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"pads", "value"}
        optional_params = set()
        super(ConstantPadding, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        return np.pad(
            x,
            self.params["pads"],
            mode="constant",
            constant_values=(self.params["value"],),
        )



from .add import Add
from .add_constant import AddConstant
from .average_pooling_2d import AveragePooling2D
from .batch_normalization import BatchNormalization
from .bias import Bias
from .bilinear_2d import Bilinear2D
from .broadcast import BroadcastTo
from .clipped_relu import ClippedReLU
from .concat import Concat
from .constant import Constant
from .convolution_2d import Convolution2D
from .cos import Cos
from .depthwise_convolution_2d import DepthwiseConvolution2D
from .div import Div
from .dropout import Dropout
from .elu import ELU
from .exp import Exp
from .function import Function
from .gemm import Gemm
from .leaky_relu import LeakyReLU
from .linear import Linear
from .local_response_normalization import LocalResponseNormalization
from .matmul import MatMul
from .max_pooling_2d import MaxPooling2D
from .mul import Mul
from .mul_constant import MulConstant
from .pad import ConstantPadding
from .relu import ReLU
from .reshape import Reshape
from .scale import Scale
from .sigmoid import Sigmoid
from .sin import Sin
from .softmax import Softmax
from .softmax_cross_entropy import SoftmaxCrossEntropy
from .sub import Sub
from .sum import Sum
from .swish import Swish
from .tan import Tan
from .tanh import Tanh
from .transpose import Transpose
from .unpooling_2d import Unpooling2D



import numpy as np

from .function import Function


class Concat(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"axis"}
        optional_params = set()
        super(Concat, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, *xs):
        return np.concatenate(xs, self.params["axis"])



import numpy as np

from .function import Function


class LocalResponseNormalization(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"n", "k", "alpha", "beta"}
        optional_params = set()
        super(LocalResponseNormalization, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        RA2 = np.square(x)
        R = RA2.copy()
        for i in range(1, self.params["n"] // 2 + 1):
            R[:, i:] += RA2[:, :-i]
            R[:, :-i] += RA2[:, i:]
        R = self.params["k"] + self.params["alpha"] * R
        R = R ** -self.params["beta"]
        R = x * R
        return R



import os
import sys

import nnoir
import numpy as np


def single_function_model(function, _inputs, _outputs, **kwargs):
    inputs = [nnoir.Value(x[0], shape=x[1], dtype="<f4") for x in _inputs]
    outputs = [nnoir.Value(x[0], shape=x[1], dtype="<f4") for x in _outputs]
    nodes = inputs + outputs
    input_names = [x.name for x in inputs]
    output_names = [x.name for x in outputs]
    functions = [getattr(nnoir.functions, function.partition("_")[0])(input_names, output_names, **kwargs)]
    actual = nnoir.NNOIR(
        function.encode(),
        b"nnoir2chainer_test",
        "0.1",
        input_names,
        output_names,
        nodes,
        functions,
    )
    expected = nnoir.load(os.path.join(os.path.abspath(os.path.dirname(__file__)), function + ".nnoir"))
    xs = [np.random.randn(*x[1]).astype("<f4") for x in _inputs]
    actuals = actual.run(*xs)
    expecteds = expected.run(*xs)
    assert len(expecteds) == len(actuals)
    for a, e in zip(actuals, expecteds):
        print(a - e)
        assert (a == e).all()


def test_Add():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10)), (b"v1", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_AddConstant():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        value=2.0,
    )


def test_AveragePooling2D():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (2, 3, 3, 3))],
        kernel=[2, 3],
        stride=[1, 2],
        pad_h=[0, 0],
        pad_w=[1, 2],
        count_exclude_pad=False,
    )


def test_BatchNormalization():
    shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", shape)],
        [(b"v2", shape)],
        eps=2e-05,
        avg_mean=np.array([6, 7, 8], dtype=np.float32),
        avg_var=np.array([9, 10, 11], dtype=np.float32),
        gamma=np.array([0, 1, 2], dtype=np.float32),
        beta=np.array([3, 4, 5], dtype=np.float32),
    )


def test_Bias():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (2, 3, 4, 5))],
        axis=1,
        b=np.arange(3).astype(np.float32),
    )


def test_Bilinear2D():
    in_shape = (2, 3, 9, 10)
    out_shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", in_shape)],
        [(b"v2", out_shape)],
        size=tuple(out_shape[2:]),
    )


def test_Bilinear2D_align_none():
    in_shape = (2, 3, 9, 10)
    out_shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", in_shape)],
        [(b"v2", out_shape)],
        size=tuple(out_shape[2:]),
        mode=b"align_none",
    )


def test_Bilinear2D_align_corners():
    in_shape = (2, 3, 9, 10)
    out_shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", in_shape)],
        [(b"v2", out_shape)],
        size=tuple(out_shape[2:]),
        mode=b"align_corners",
    )


def test_Bilinear2D_align_centers():
    in_shape = (2, 3, 9, 10)
    out_shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", in_shape)],
        [(b"v2", out_shape)],
        size=tuple(out_shape[2:]),
        mode=b"align_centers",
    )


def test_BroadcastTo():
    in_shape = (1, 1, 4, 5)
    out_shape = (2, 3, 4, 5)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", in_shape)],
        [(b"v2", out_shape)],
        shape=out_shape,
    )


def test_ClippedReLU():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        upper=40.0,
    )


def test_Concat():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 11)), (b"v1", (10, 12))],
        [(b"v2", (10, 10))],
        axis=1,
    )


def test_ConstantPadding():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (3, 5, 7, 10))],
        pads=((1, 0), (1, 1), (1, 2), (0, 5)),
        value=1.0,
    )


def test_Convolution2D():
    batch = 2
    in_ch = 4
    in_h = 10
    in_w = 9
    out_ch = 7
    out_h = 6
    out_w = 3
    kh = 4
    kw = 3
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (batch, in_ch, in_h, in_w))],
        [(b"v1", (batch, out_ch, out_h, out_w))],
        W=np.arange(out_ch * in_ch * kh * kw).reshape(out_ch, in_ch, kh, kw).astype(np.float32),
        b=np.arange(out_ch).astype(np.float32),
        pad_h=(2, 2),
        pad_w=(1, 1),
        stride=(2, 3),
        dilate=(1, 1),
        groups=1,
    )


def test_Cos():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_DepthwiseConvolution2D():
    sy, sx = 2, 3
    ph, pw = 1, 3
    kh, kw = 3, 4
    batch = 2
    in_ch = 6
    in_h = 10
    in_w = 9
    ch_mul = 2
    out_ch = in_ch * ch_mul
    out_h = 5
    out_w = 4
    dy, dx = (2, 3)
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (batch, in_ch, in_h, in_w))],
        [(b"v1", (batch, out_ch, out_h, out_w))],
        W=np.arange(ch_mul * in_ch * kh * kw).reshape(ch_mul, in_ch, kh, kw).astype(np.float32),
        b=np.arange(out_ch).astype(np.float32),
        pad_h=(1, 1),
        pad_w=(3, 3),
        stride=(sy, sx),
        dilate=(dy, dx),
    )


def test_Div():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10)), (b"v1", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Dropout():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_ELU():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        alpha=0.5,
    )


def test_Exp():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v1", (1, 3, 4, 5))],
        [(b"v0", (1, 3, 4, 5))],
    )


def test_Gemm():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (3, 4)), (b"v1", (4, 5))],
        [(b"v2", (3, 5))],
        transA=False,
        transB=False,
        alpha=2.0,
        beta=1.0,
    )


def test_LeakyReLU():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        slope=0.5,
    )


def test_Linear():
    batch = 2
    in_ch = 3
    out_ch = 4
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (batch, in_ch))],
        [(b"v2", (batch, out_ch))],
        W=np.arange(out_ch * in_ch).reshape(out_ch, in_ch).astype(np.float32),
        b=np.arange(out_ch).astype(np.float32),
    )


def test_LocalResponseNormalization():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        n=6,
        k=3.0,
        alpha=0.0002,
        beta=0.8,
    )


def test_MatMul():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (3, 4)), (b"v1", (4, 5))],
        [(b"v2", (3, 5))],
    )


def test_MaxPooling2D():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (2, 3, 5, 4))],
        kernel=[2, 3],
        stride=[1, 2],
        pad_h=[1, 1],
        pad_w=[2, 3],
    )


def test_Mul():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10)), (b"v1", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_MulConstant():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        value=2.0,
    )


def test_ReLU():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Reshape():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (5, 20))],
        shape=(5, 20),
    )


def test_Scale():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (2, 3, 4, 5))],
        axis=1,
        W=np.arange(3).astype(np.float32),
        b=np.arange(3).astype(np.float32),
    )


def test_Sigmoid():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Sin():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Softmax():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
        axis=1,
    )


def test_Sub():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10)), (b"v1", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Sum():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (2, 1, 4, 1))],
        axes=(1, 3),
        keepdims=True,
    )


def test_Swish():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10, 3))],
        [(b"v2", (10, 10, 3))],
        beta=np.arange(10).astype(np.float32),
    )


def test_Tan():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (10, 10))],
        [(b"v2", (10, 10))],
    )


def test_Transpose():
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (2, 3, 4, 5))],
        [(b"v2", (4, 5, 2, 3))],
        axes=(2, 3, 0, 1),
    )


def test_Unpooling2D():
    kh, kw = 2, 3
    sy, sx = 1, 2
    batch = 2
    ch = 3
    in_h, in_w = 5, 6
    out_h, out_w = 4, 9
    ph, pw = 1, 2
    single_function_model(
        sys._getframe().f_code.co_name[5:],
        [(b"v0", (batch, ch, in_h, in_w))],
        [(b"v2", (batch, ch, out_h, out_w))],
        kh=kh,
        kw=kw,
        sy=sy,
        sx=sx,
        ph=ph,
        pw=pw,
        cover_all=False,
        outh=out_h,
        outw=out_w,
    )



import numpy as np

from .function import Function


class Gemm(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"transA", "transB", "alpha", "beta"}
        optional_params = set()
        super(Gemm, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, a, b, c=None):
        a = a if not self.params["transA"] else a.T
        b = b if not self.params["transB"] else b.T
        result = self.params["alpha"] * np.dot(a, b)
        if c is not None:
            result += self.params["beta"] * c
        return result



from .function import Function


class Reshape(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"shape"}
        optional_params = set()
        super(Reshape, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        return x.reshape(self.params["shape"])



import numpy as np

from .function import Function


class ELU(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"alpha"}
        optional_params = set()
        super(ELU, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        R = x.copy()
        v = R < 0
        R[v] = self.params["alpha"] * (np.exp(R[v]) - 1)
        return R



from .function import Function


class SoftmaxCrossEntropy(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"normalize", "cache_score"}
        optional_params = set()
        super(SoftmaxCrossEntropy, self).__init__(inputs, outputs, params, required_params, optional_params)



import re

import msgpack


class InvalidNNOIRData(Exception):
    def __init__(self, message):
        self.message = message


class NNOIR:
    def __init__(
        self,
        name,
        generator_name,
        generator_version,
        inputs,
        outputs,
        values,
        functions,
    ):
        cident = re.compile(rb"^[_A-Za-z][_0-9A-Za-z]*$")
        c_keywords = [
            "auto",
            "break",
            "case",
            "char",
            "const",
            "continue",
            "default",
            "do",
            "double",
            "else",
            "enum",
            "extern",
            "float",
            "for",
            "goto",
            "if",
            "inline",
            "int",
            "long",
            "register",
            "restrict",
            "return",
            "short",
            "signed",
            "sizeof",
            "static",
            "struct",
            "switch",
            "typedef",
            "union",
            "unsigned",
            "void",
            "volatile",
            "while",
            "_Alignas",
            "_Alignof",
            "_Atomic",
            "_Bool",
            "_Complex",
            "_Generic",
            "_Imaginary",
            "_Noreturn",
            "_Static_assert",
            "_Thread_local",
        ]
        if not cident.match(name) or name.decode() in c_keywords:
            raise InvalidNNOIRData('graph name "{}" MUST be C identifier.'.format(name))
        self.name = name
        self.generator_name = generator_name
        self.generator_version = generator_version
        self.inputs = inputs
        self.outputs = outputs
        self.functions = functions
        self.values = values
        vident = re.compile(rb"v[_0-9A-Za-z]*")  # 'v' prefixed
        for v in self.values:
            if not vident.match(v.name) or v.name.decode() in c_keywords:
                raise InvalidNNOIRData('value name "{}" MUST be "v" prefixed C identifier.'.format(v.name))

    def to_model(self):
        return {
            b"name": self.name,
            b"generator": {
                b"name": self.generator_name,
                b"version": self.generator_version,
            },
            b"inputs": self.inputs,
            b"outputs": self.outputs,
            b"values": [v.dump() for v in self.values],
            b"functions": [f.dump() for f in self.functions],
        }

    def to_nnoir(self):
        return {b"nnoir": {b"version": 0, b"model": self.to_model()}}

    def pack(self):
        return msgpack.packb(self.to_nnoir(), use_bin_type=False)

    def dump(self, file_name):
        result = self.pack()
        with open(file_name, "w") as f:
            f.buffer.write(result)

    def run(self, *inputs):
        env = dict(zip(self.inputs, inputs))

        def cast(vs):
            res = []
            for (n, arr) in vs:
                v = [v for v in self.values if v.name == n][0]
                res.append((n, arr.astype(v.dtype)))
            return res

        def eval(n):
            if n not in env:
                for f in self.functions:
                    if n not in f.outputs:
                        continue
                    inputs = [eval(i) for i in f.inputs]
                    outputs = f.run(*inputs)
                    if type(outputs) == list:
                        pass
                    elif type(outputs) == tuple:
                        outputs = list(outputs)
                    else:
                        outputs = [outputs]
                    env.update(dict(cast(zip(f.outputs, outputs))))
                    break
            return env[n]

        result = [eval(o) for o in self.outputs]
        del cast
        del eval
        return result



import numpy as np

from .function import Function


class Swish(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"beta"}
        optional_params = set()
        super(Swish, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        beta = self.params["beta"]
        d = len(x.shape) - 1 - len(beta.shape)
        beta = beta.reshape((1,) + beta.shape + (1,) * d)
        return x * (0.5 * np.tanh(0.5 * beta * x) + 0.5)



import numpy as np

from . import util
from .function import Function


class Unpooling2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {
            "kh",
            "kw",
            "sy",
            "sx",
            "ph",
            "pw",
            "cover_all",
            "outh",
            "outw",
        }
        optional_params = set()
        super(Unpooling2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        col = np.tile(x[:, :, None, None], (1, 1, self.params["kh"], self.params["kw"], 1, 1))
        R = util.col2im_cpu(
            col,
            (self.params["sy"], self.params["sx"]),
            self.params["ph"],
            self.params["pw"],
            self.params["outh"],
            self.params["outw"],
        )
        return R



import io

import msgpack
import numpy as np


class NNTC:
    def __init__(self, model_name, inputs, outputs):
        self.version = 0
        self.model = model_name
        self.test_case = TestCase(inputs, outputs)

    def to_dict(self):
        return {
            b"version": self.version,
            b"model": self.model,
            b"test_case": self.test_case.to_dict(),
        }

    def pack(self):
        return msgpack.packb(self.to_dict(), use_bin_type=False)

    def dump(self, file_name):
        result = self.pack()
        with open(file_name, "w") as f:
            f.buffer.write(result)


class TestCase:
    def __init__(self, inputs, outputs):
        self.inputs = inputs
        self.outputs = outputs

    def to_dict(self):
        return {
            b"inputs": [i.to_dict() for i in self.inputs],
            b"outputs": [o.to_dict() for o in self.outputs],
        }


class TestValue:
    def __init__(self, name, ndarray):
        self.name = name
        self.ndarray = ndarray

    def to_dict(self):
        def encode_ndarray(obj):
            x = None
            with io.BytesIO() as out:
                np.save(out, obj.copy())
                x = out.getvalue()
            return x

        return {b"value_name": self.name, b"ndarray": encode_ndarray(self.ndarray)}



from . import util
from .function import Function


class Constant(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"value"}
        optional_params = set()
        super(Constant, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self):
        return self.params["value"]



from . import util
from .function import Function


class MaxPooling2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"kernel", "stride", "pad_h", "pad_w"}
        optional_params = set()
        super(MaxPooling2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        img, col = util.im2col_cpu(
            x,
            self.params["kernel"],
            self.params["stride"],
            self.params["pad_h"],
            self.params["pad_w"],
            pval=-float("inf"),
        )
        R = col.max(axis=(2, 3))
        return R



import numpy as np

from .function import Function


class Bilinear2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"size"}
        optional_params = {"mode"}
        super(Bilinear2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        out_h = self.params["size"][0]
        out_w = self.params["size"][1]
        batch, ch, in_h, in_w = x.shape
        if "mode" in self.params and self.params["mode"] == b"align_none":
            # Mode: TF1(default), onnxruntime
            h_scale = float(in_h) / float(out_h)
            w_scale = float(in_w) / float(out_w)
            V = np.minimum(np.arange(out_h) * h_scale, in_h - 1).reshape(out_h, 1)
            V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
            U = np.minimum(np.arange(out_w) * w_scale, in_w - 1).reshape(1, out_w)
            U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
        elif "mode" not in self.params or self.params["mode"] == b"align_corners":
            # Mode: TF1(align_corners), Chainer, PyTorch
            h_scale = float(in_h - 1) / float(out_h - 1)
            w_scale = float(in_w - 1) / float(out_w - 1)
            V = (np.arange(out_h) * h_scale).reshape(out_h, 1)
            V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
            U = (np.arange(out_w) * w_scale).reshape(1, out_w)
            U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
        elif "mode" in self.params and self.params["mode"] == b"align_centers":
            # Mode: TF2, OpenCV
            h_scale = float(in_h) / float(out_h)
            w_scale = float(in_w) / float(out_w)
            V = np.minimum(
                np.maximum((np.arange(out_h) + 0.5) * h_scale - 0.5, 0.0),
                float(in_h - 1),
            ).reshape(out_h, 1)
            V0 = np.maximum(0, np.minimum(np.floor(V), in_h - 2)).astype("int32")
            U = np.minimum(
                np.maximum((np.arange(out_w) + 0.5) * w_scale - 0.5, 0.0),
                float(in_w - 1),
            ).reshape(1, out_w)
            U0 = np.maximum(0, np.minimum(np.floor(U), in_w - 2)).astype("int32")
        else:
            raise Exception("unknown Bilinear mode")
        V1 = V0 + 1
        U1 = U0 + 1
        W00 = (U1 - U) * (V1 - V)
        W01 = (U - U0) * (V1 - V)
        W10 = (U1 - U) * (V - V0)
        W11 = (U - U0) * (V - V0)
        V0 = V0.reshape(out_h)
        V1 = np.minimum(V1.reshape(out_h), in_h - 1)
        U0 = U0.reshape(out_w)
        U1 = np.minimum(U1.reshape(out_w), in_w - 1)
        x0 = x[:, :, V0, :]
        x1 = x[:, :, V1, :]
        result = W00 * x0[:, :, :, U0] + W01 * x0[:, :, :, U1] + W10 * x1[:, :, :, U0] + W11 * x1[:, :, :, U1]
        return result



import numpy as np

from . import util
from .function import Function


class Linear(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"W", "b"}  # (out_ch, in_ch)
        optional_params = {"y_scale", "y_zero", "w_scale", "w_zero"}
        super(Linear, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        x_ = x.reshape(x.shape[0], -1)
        w_ = np.transpose(self.params["W"])
        if self.params["W"].dtype == np.uint8:
            func = lambda x, w: x.dot(w)
            result = util.calc_with_uint8_weight(func, x_, w_, self.params["w_scale"], self.params["w_zero"])
        else:
            assert self.params["W"].dtype == np.float32
            result = x_.dot(w_)
        if self.params["b"] is not None:
            assert self.params["b"].dtype == np.float32
            result += self.params["b"]
        return result



import numpy as np

from . import util
from .function import Function


class Sum(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"axes", "keepdims"}
        optional_params = set()
        super(Sum, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        return np.sum(x, axis=tuple(self.params["axes"]), keepdims=self.params["keepdims"])



import numpy as np

from . import util
from .function import Function


class DepthwiseConvolution2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"W", "b", "stride", "pad_h", "pad_w", "dilate"}
        optional_params = {"y_scale", "y_zero", "w_scale", "w_zero"}
        super(DepthwiseConvolution2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        W = self.params["W"]
        b = self.params["b"]
        kernel = (self.params["W"].shape[2], self.params["W"].shape[3])
        img, col = util.im2col_cpu(
            x,
            kernel,
            self.params["stride"],
            self.params["pad_h"],
            self.params["pad_w"],
            self.params["dilate"],
        )
        B, C, KY, KX, IY, IX = col.shape
        D = W.shape[0]  # (D, C, KY, KX)
        c_ = col.transpose(1, 0, 4, 5, 2, 3).reshape((C, B * IY * IX, KY * KX))
        w_ = W.transpose(1, 2, 3, 0).reshape((C, KY * KX, D))
        if self.params["W"].dtype == np.uint8:
            func = lambda x, w: np.matmul(x, w).astype(col.dtype, copy=False)
            y = util.calc_with_uint8_weight(func, c_, w_, self.params["w_scale"], self.params["w_zero"])
        else:
            assert self.params["W"].dtype == np.float32
            y = np.matmul(c_, w_).astype(col.dtype, copy=False)
        y = y.reshape((C, B, IY * IX, D)).transpose(1, 0, 3, 2).reshape((B, C * D, IY, IX))
        if b is not None:
            assert self.params["W"].dtype == np.float32
            y += b[None, :, None, None]
        return y



import io

import msgpack
import numpy

from .functions import *
from .nnoir import NNOIR
from .value import Value


def load(nnoir_file):
    with open(nnoir_file, "rb") as f:
        nnoir = msgpack.unpackb(f.read(), raw=True)
    name = nnoir[b"nnoir"][b"model"][b"name"]
    generator_name = nnoir[b"nnoir"][b"model"][b"generator"][b"name"]
    generator_version = nnoir[b"nnoir"][b"model"][b"generator"][b"version"]
    inputs = nnoir[b"nnoir"][b"model"][b"inputs"]
    outputs = nnoir[b"nnoir"][b"model"][b"outputs"]
    vs = [_decode_value(v) for v in nnoir[b"nnoir"][b"model"][b"values"]]
    fs = [_decode_function(f) for f in nnoir[b"nnoir"][b"model"][b"functions"]]
    return NNOIR(name, generator_name, generator_version, inputs, outputs, vs, fs)


def _decode_function(function):
    inputs = function[b"inputs"]
    outputs = function[b"outputs"]
    params = {}
    for k, v in function[b"params"].items():
        if type(v) is dict:
            params[k.decode()] = numpy.load(io.BytesIO(v[b"ndarray"]))
        else:
            params[k.decode()] = v
    name = function[b"name"].decode()
    return globals()[name](inputs, outputs, **params)


def _decode_value(value):
    return Value(value[b"name"], np_array=None, dtype=value[b"dtype"], shape=value[b"shape"])



from .function import Function


class Div(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = set()
        optional_params = set()
        super(Div, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x1, x2):
        return x1 / x2



import numpy as np

from .function import Function


class Bias(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"axis", "b"}
        optional_params = set()
        super(Bias, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        shape_post_len = len(x.shape) - self.params["axis"] - len(self.params["b"].shape)
        shape = (1,) * self.params["axis"] + self.params["b"].shape + (1,) * shape_post_len
        return x + np.reshape(self.params["b"], shape)



import numpy as np


def get_conv_outsize(size, k, s, pre_p, post_p, cover_all=False, d=1):
    dk = k + (k - 1) * (d - 1)
    return (size + pre_p + post_p - dk) // s + 1


def im2col_cpu(img, kernel, stride, ph, pw, dilate=(1, 1), pval=0, cover_all=False):
    n, c, h, w = img.shape
    kh, kw = kernel
    sy, sx = stride
    dy, dx = dilate
    pre_ph = ph[0]
    post_ph = ph[1]
    pre_pw = pw[0]
    post_pw = pw[1]
    out_h = get_conv_outsize(h, kh, sy, pre_ph, post_ph, cover_all, dy)
    out_w = get_conv_outsize(w, kw, sx, pre_pw, post_pw, cover_all, dx)

    img = np.pad(
        img,
        ((0, 0), (0, 0), (pre_ph, post_ph), (pre_pw, post_pw)),
        mode="constant",
        constant_values=(pval,),
    )
    col = np.ndarray((n, c, kh, kw, out_h, out_w), dtype=img.dtype)

    for j in range(kh):
        jdy = j * dy
        j_lim = jdy + sy * out_h
        for i in range(kw):
            idx = i * dx
            i_lim = idx + sx * out_w
            col[:, :, j, i, :, :] = img[:, :, jdy:j_lim:sy, idx:i_lim:sx]

    return img, col


def col2im_cpu(col, stride, ph, pw, h, w, dy=1, dx=1):
    sy, sx = stride
    n, c, kh, kw, out_h, out_w = col.shape
    img = np.zeros((n, c, h + 2 * ph + sy - 1, w + 2 * pw + sx - 1), dtype=col.dtype)
    for j in range(kh):
        jdy = j * dy
        j_lim = jdy + sy * out_h
        for i in range(kw):
            idx = i * dx
            i_lim = idx + sx * out_w
            img[:, :, jdy:j_lim:sy, idx:i_lim:sx] += col[:, :, j, i]
    return img[:, :, ph : h + ph, pw : w + pw]


def get_quantization_factors(_min, _max):
    _max = max(_max, 0)
    _min = min(_min, 0)
    scale = ((_max - _min) / 255).astype("float32")
    zero = -_min / scale
    zero = np.round(zero.clip(0, 255)).astype("uint8")
    return np.asscalar(scale), np.asscalar(zero)


def quantize_variable(x):
    scale, zero = get_quantization_factors(x.astype("float32").min(), x.astype("float32").max())
    quantized_x = np.round((x / scale + zero).clip(0, 255)).astype("uint8")
    return quantized_x, scale, zero


def calc_with_uint8_weight(func, x_fp32, w_uint8, w_scale, w_zero):
    x_uint8, x_scale, x_zero = quantize_variable(x_fp32)
    x = x_uint8.astype("int32") - int(x_zero)
    w = w_uint8.astype("int32") - int(w_zero)
    y = func(x, w).astype("float32") * x_scale * w_scale
    return y



import numpy as np

from .function import Function


class Tanh(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = set()
        optional_params = set()
        super(Tanh, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        R = x.copy()
        return np.tanh(R)



import io

import nnoir
import numpy


class Function(object):
    def __init__(self, inputs, outputs, params, required_params, optional_params):
        if required_params - set(params.keys()) != set():
            lacks = ", ".join(required_params - set(params.keys()))
            raise Exception("lack of required parameter: {}".format(lacks))
        if set(params.keys()) - required_params - optional_params != set():
            unknowns = ", ".join(set(params.keys()) - required_params - optional_params)
            raise Exception("unknown parameter: {}".format(unknowns))
        self.inputs = inputs
        self.outputs = outputs
        self.params = params

    def dump(self):
        def encode_ndarray(obj):
            x = None
            with io.BytesIO() as out:
                numpy.save(out, obj.copy())
                x = out.getvalue()
            return {b"ndarray": x}

        binary_params = {}
        for k, v in self.params.items():
            if type(v) is numpy.ndarray:
                binary_params[k.encode()] = encode_ndarray(v)
            elif type(v) is nnoir.NNOIR:
                binary_params[k.encode()] = v.to_model()
            else:
                binary_params[k.encode()] = v
        return {
            b"name": self.__class__.__name__,
            b"inputs": self.inputs,
            b"outputs": self.outputs,
            b"params": binary_params,
        }



import numpy as np

from . import util
from .function import Function


class Convolution2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"W", "b", "pad_h", "pad_w", "stride", "dilate", "groups"}
        optional_params = {"y_scale", "y_zero", "w_scale", "w_zero"}
        super(Convolution2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def grouped_convolution(self, x):
        G = self.params["groups"]
        batch, in_ch, in_h, in_w = x.shape
        out_ch, _, kh, kw = self.params["W"].shape
        icg = in_ch // G
        ocg = out_ch // G

        kernel = (self.params["W"].shape[2], self.params["W"].shape[3])
        _, x = util.im2col_cpu(
            x,
            kernel,
            self.params["stride"],
            self.params["pad_h"],
            self.params["pad_w"],
            self.params["dilate"],
        )
        out_h, out_w = x.shape[-2:]

        x = x.transpose(1, 2, 3, 0, 4, 5)  # (in_ch, kh, kw, batch, out_h, out_w)
        x = x.reshape(G, icg * kh * kw, batch * out_h * out_w)

        W = self.params["W"].reshape(G, ocg, icg * kh * kw)

        if self.params["W"].dtype == np.uint8:
            func = lambda x, w: np.matmul(w, x).astype(x.dtype, copy=False)
            y = util.calc_with_uint8_weight(func, x, W, self.params["w_scale"], self.params["w_zero"])
        else:
            assert W.dtype == np.float32
            y = np.matmul(W, x).astype(x.dtype, copy=False)
        y = y.reshape(out_ch, batch, out_h, out_w)
        y = y.transpose(1, 0, 2, 3)
        if self.params["b"] is not None:
            assert self.params["b"].dtype == np.float32
            y += self.params["b"].reshape(1, self.params["b"].size, 1, 1)

        return (y,)

    def run(self, x):
        if self.params["groups"] > 1:
            return self.grouped_convolution(x)

        kernel = (self.params["W"].shape[2], self.params["W"].shape[3])
        img, col = util.im2col_cpu(
            x,
            kernel,
            self.params["stride"],
            self.params["pad_h"],
            self.params["pad_w"],
            self.params["dilate"],
        )
        if self.params["W"].dtype == np.uint8:
            func = lambda x, w: np.tensordot(x, w, ((1, 2, 3), (1, 2, 3)))
            R = util.calc_with_uint8_weight(func, col, self.params["W"], self.params["w_scale"], self.params["w_zero"])
        else:
            assert self.params["W"].dtype == np.float32
            R = np.tensordot(col, self.params["W"], ((1, 2, 3), (1, 2, 3)))
        if self.params["b"] is not None:
            assert self.params["b"].dtype == np.float32
            R += self.params["b"]
        R = np.rollaxis(R, 3, 1)
        return R



import os

from setuptools import find_packages, setup

exec(open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "nnoir", "_version.py")).read())

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as f:
    long_description = f.read()

setup(
    name="nnoir",
    version=__version__,
    description="API for NNOIR",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Idein/nnoir/tree/master/nnoir",
    author="Idein Inc.",
    author_email="fujii@idein.jp",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="nnoir machine learning",
    packages=find_packages(),
    install_requires=["numpy", "msgpack"],
    scripts=["nnoir2dot", "nnrunner"],
)



class Value:
    def __init__(self, name, np_array=None, dtype=None, shape=None):
        self.name = name
        if np_array is None:
            self.dtype = dtype
            self.shape = shape
        else:
            self.dtype = np_array.dtype.str.encode()
            self.shape = np_array.shape

    def dump(self):
        return {b"name": self.name, b"dtype": self.dtype, b"shape": self.shape}



import numpy as np

from .function import Function


class BatchNormalization(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"eps", "avg_mean", "avg_var", "gamma", "beta"}
        optional_params = set()
        super(BatchNormalization, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        if self.params["gamma"] is None:
            self.params["gamma"] = np.ones(self.params["avg_mean"].size).astype(np.float32)
        if self.params["beta"] is None:
            self.params["beta"] = np.zeros(self.params["avg_mean"].size).astype(np.float32)

        shape = (1, self.params["gamma"].size) + ((1,) * (len(x.shape) - 2))
        gamma = self.params["gamma"].reshape(shape)
        beta = self.params["beta"].reshape(shape)
        avg_mean = self.params["avg_mean"].reshape(shape)
        avg_var = self.params["avg_var"].reshape(shape)
        return gamma * (x - avg_mean) / np.sqrt(avg_var + self.params["eps"]) + beta



import numpy as np

from . import util
from .function import Function


class AveragePooling2D(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"kernel", "stride", "pad_h", "pad_w", "count_exclude_pad"}
        optional_params = set()
        super(AveragePooling2D, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        if self.params["count_exclude_pad"]:
            img, col = util.im2col_cpu(
                x,
                self.params["kernel"],
                self.params["stride"],
                self.params["pad_h"],
                self.params["pad_w"],
            )
            _, mask = util.im2col_cpu(
                np.ones(x.shape, dtype=np.int32),
                self.params["kernel"],
                self.params["stride"],
                self.params["pad_h"],
                self.params["pad_w"],
            )
            return col.sum(axis=(2, 3)) / mask.sum(axis=(2, 3))
        else:
            img, col = util.im2col_cpu(
                x,
                self.params["kernel"],
                self.params["stride"],
                self.params["pad_h"],
                self.params["pad_w"],
            )
            return col.mean(axis=(2, 3))



from .function import Function


class ReLU(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = set()
        optional_params = set()
        super(ReLU, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        R = x.copy()
        R[R < 0] = 0
        return R



from .function import Function


class Transpose(Function):
    def __init__(self, inputs, outputs, **params):
        required_params = {"axes"}
        optional_params = set()
        super(Transpose, self).__init__(inputs, outputs, params, required_params, optional_params)

    def run(self, x):
        return x.transpose(self.params["axes"])
